#include "stdafx.h"

#include <algorithm>
#include <deque>
#include <vector>

#include "character.h"
#include "ff3se.h"
#include "tile.h"

Character character[numCharacters];
RGBTRIPLE charPals[32][16];

std::vector<BYTE> charTiles;
USHORT fullOffset[numFullPics * 6];
USHORT walkWave2PoseOffset[(numWalkPics + 4) * 6];
USHORT operaOffset[(numWalkPics + 3) * 6];
USHORT walkPoseOffset[(numWalkPics + 1) * 6];
USHORT walkOffset[numWalkPics * 6];

/*************************************/
Character::Character(void)
{
	type = FULL;

	fixedPalette = false;
}

/*************************************/
Character::~Character(void)
{
}

/*************************************/
//Changes the Portrait's palette and updates the bitmap
void Character::changePalette(BYTE p)
{
	int i, j;

	if (fixedPalette)
		return;

	palette = p;

	for (i = 0; i < getNumPics(); i++)
	{
		for (j = 0; j < 16; j++)
		{
			pics[i].setColor(j, charPals[palette][j]);
		}
	}
}

/*************************************/
//Clears all sprite sfor this character
void Character::clear(void)
{
	int i, j;
	int num = getNumPics();

	if (num == numFullPics)
	{
		pics[numFullPics - 1].createBlank(getPixelHeight(), getPixelWidth());
		num--;
	}
	for (i = 0; i < num; i++)
	{
		pics[i].createBlank(getPixelWidth(), getPixelHeight());
		for (j = 0; j < 16; j++)
		{
			pics[i].setColor(j, charPals[palette][j]);
		}
	}
}

/*************************************/
//Copies a sprite from another character
//If copyAll is true copies all sprites
void Character::copyImage(int num, int copyAll)
{
	int i;

	if (copyAll == -1)
	{
		tiles = character[num].tiles;
		palette = character[num].palette;
	}
	else
	{
		const USHORT *offsets = getTileOffsets();
		int tilesPerImage = getTileHeight() * getTileWidth();
		for (i = 0; i < tilesPerImage; i++)
		{
			int offset = offsets[copyAll * tilesPerImage + i] * 2;
			auto start = character[num].tiles.cbegin() + offset;
			auto end = start + 64;
			std::copy(start, end, tiles.begin() + offset);
		}
	}

	createBitmaps();
}

// 0xD092
USHORT animated_offsets[] = {
	0x0000, 0x0000, 0x0020, 0x0040, 0x0060, 0x0080,
	0x0000, 0x0000, 0x00A0, 0x00C0, 0x00E0, 0x0100,
	0x0000, 0x0000, 0x0120,	0x0140, 0x0160, 0x0180,
	0x0000, 0x0000, 0x01A0, 0x01C0,	0x01E0, 0x0200,
};

USHORT offsets_16[] = {
	0x0000, 0x0020, 0x0040, 0x0060,
	0x0080, 0x00A0, 0x00C0, 0x00E0,
	0x0100, 0x0120, 0x0140, 0x0160,
	0x0180, 0x01A0, 0x01C0, 0x01E0,
};

USHORT offsets_32[] = {
	0x0000, 0x0020, 0x0040, 0x0060, 0x0080, 0x00A0, 0x00C0, 0x00E0,
	0x0100, 0x0120, 0x0140, 0x0160,	0x0180, 0x01A0, 0x01C0, 0x01E0,
	0x0200, 0x0220, 0x0240, 0x0260, 0x0280, 0x02A0, 0x02C0, 0x02E0,
	0x0300, 0x0320, 0x0340, 0x0360,	0x0380, 0x03A0, 0x03C0, 0x03E0,
};

 // 0xD07A
USHORT offsets_ramuh[] = {
	0x0000, 0x0020, 0x0040, 0x0060, 0x0080, 0x00A0,
	0x00C0, 0x0020, 0x00E0, 0x0060, 0x0100, 0x00A0,
	0x0120, 0x0140, 0x0040, 0x0060, 0x0080, 0x00A0,
};

USHORT offsets_guard_riding[] = {
	0x0000, 0x0020, 0x0040, 0x0060, 0x0080, 0x00A0,
	0x00C0, 0x00E0, 0x0100, 0x0120, 0x0080, 0x00A0,
};

const USHORT *Character::getTileOffsets() const {
	switch (type)
	{
		case FULL:
			return fullOffset;
		case WALK_WAVE_2_POSE:
			return walkWave2PoseOffset;
		case OPERA:
			return operaOffset;
		case WALK_POSE:
			return walkPoseOffset;
		case WALK:
			return walkOffset;
		case RAMUH:
			return offsets_ramuh;

		case GUARD_RIDING:
			return offsets_guard_riding;

		case ANIMATED:
			return animated_offsets;

		case STATIC_16:
			return offsets_16;

		case STATIC_32:
			return offsets_32;

		default:
			return fullOffset;
	}
}

/*************************************/
//Creates all the Bitmaps for the Map Sprite
void Character::createBitmaps(void)
{
	int i, j;
	int x, y;

	if (fixedPalette)		//Set up palette for Esper Terra
	{
		for (i = 0; i < numFullPics; i++)
		{
			pics[i].createBlank(getPixelWidth(), getPixelHeight());
			pics[i].setColor(0, makeRGBTRIPLE(31 * 8, 5 * 8, 0));
			pics[i].setColor(1, makeRGBTRIPLE(11 * 8, 2 * 8, 8 * 8));
			pics[i].setColor(2, makeRGBTRIPLE(31 * 8, 31 * 8, 27 * 8));
			pics[i].setColor(3, makeRGBTRIPLE(18 * 8, 0, 0));
			pics[i].setColor(4, makeRGBTRIPLE(31 * 8, 31 * 8, 31 * 8));
			pics[i].setColor(5, makeRGBTRIPLE(25 * 8, 16 * 8, 31 * 8));
			pics[i].setColor(6, makeRGBTRIPLE(31 * 8, 31 * 8, 31 * 8));
			pics[i].setColor(7, makeRGBTRIPLE(17 * 8, 14 * 8, 17 * 8));
			pics[i].setColor(8, makeRGBTRIPLE(23 * 8, 18 * 8, 0));
			pics[i].setColor(9, makeRGBTRIPLE(18 * 8, 0, 0));
			pics[i].setColor(10, makeRGBTRIPLE(25 * 8, 16 * 8, 31 * 8));
			pics[i].setColor(11, makeRGBTRIPLE(16 * 8, 11 * 8, 27 * 8));
			pics[i].setColor(12, makeRGBTRIPLE(20 * 8, 5 * 8, 4 * 8));
			pics[i].setColor(13, makeRGBTRIPLE(30 * 8, 16 * 8, 10 * 8));
			pics[i].setColor(14, makeRGBTRIPLE(14 * 8, 7 * 8, 0));
			pics[i].setColor(15, makeRGBTRIPLE(18 * 8, 15 * 8, 10 * 8));
		}
	}
	else
	{
		int num_pics = getNumPics();

		if (num_pics == numFullPics)
		{
			pics[num_pics - 1].createBlank(getPixelHeight(), getPixelWidth());
			for (j = 0; j < 16; j++)
			{
				pics[num_pics - 1].setColor(j, charPals[palette][j]);
			}
			num_pics--;
			if (type == NO_RIDING)
			{
				num_pics = numPreRidingPics;
			}
		}

		for (i = 0; i < num_pics; i++)
		{
			pics[i].createBlank(getPixelWidth(), getPixelHeight());
			for (j = 0; j < 16; j++)
			{
				pics[i].setColor(j, charPals[palette][j]);
			}
		}
	}

	const USHORT *offset = getTileOffsets();

	int num_pics = getNumPics();

	if (num_pics == numFullPics)
	{
		for (y = 0; y < getTileWidth(); y++)
		{
			for (x = 0; x < getTileHeight(); x++)
			{
				pics[num_pics - 1].copyTile(&tiles[offset[(num_pics - 1) * getTileHeight() * getTileWidth() + y * getTileHeight() + x] * 2], x, y);
				int v = (num_pics - 1) * getTileHeight() * getTileWidth() + y * getTileHeight() + x;
				USHORT q = offset[(num_pics - 1) * getTileHeight() * getTileWidth() + y * getTileHeight() + x];
			}
		}

		num_pics--;
		if (type == NO_RIDING)
		{
			num_pics = numPreRidingPics;
		}
	}

	for (i = 0; i < num_pics; i++)
	{
		for (y = 0; y < getTileHeight(); y++)
		{
			for (x = 0; x < getTileWidth(); x++)
			{
				pics[i].copyTile(&tiles[offset[i * getTileHeight() * getTileWidth() + y * getTileWidth() + x] * 2], x, y);
			}
		}
	}
}

/*************************************/
void Character::flipImage(int flip)
{
	int i;

	for (i = 0; i < getNumPics(); i++)
	{
		pics[i].flipImage(flip);
	}
}

/*************************************/
//Loads the graphics for this map sprite
void Character::load(int handle, BYTE num)
{
	int address;
	int next_address;
	BYTE c;
	BYTE *ptr;
	int i;

	if (num <= 15)
		type = FULL;
	else if (num <= 21)
		type = NO_RIDING;
	else if (num < 26)
		type = WALK_WAVE_2_POSE;
	else if (num < 29)
		type = OPERA;
	else if (num < 49)
		type = WALK_POSE;
	else if (num < 63)
		type = WALK;
	else if (num == 63)
		type = RAMUH;
	else if (num == 64)
		type = GUARD_RIDING;
	else if (num >= 65 && num < 87)
		type = STATIC_NPC;
	else if (num >= 87 && num < 116)
		type = ANIMATED;
	else
		// check if STATIC_32 later
		type = STATIC_16;

	if (num < 22)
	{
		//Get palette number
		_lseek(handle, headerAdjust(0x2D02B) + num, SEEK_SET);
		_read(handle, &palette, 1);
		if (palette > 8)
			palette = 0;
	}
	else
	{
		palette = 0;
	}

	if (num == 18)			//Fix the palette for the Esper Terra
		fixedPalette = true;
	else
		fixedPalette = false;

	//Load tiles
	_lseek(handle, headerAdjust(0x00D2F2) + num * 2, SEEK_SET);
	_read(handle, &c, 1);
	address = c;
	_read(handle, &c, 1);
	address += c * 256;

	_lseek(handle, headerAdjust(0x00D43C) + num * 2, SEEK_SET);
	_read(handle, &c, 1);

	address += c * 65536;
	address = mem2ROM(address);
	address -= headerAdjust(0x150200);

	if (num == numCharacters - 1)
	{
		size = 0x100;
	}
	else
	{
		_lseek(handle, headerAdjust(0x00D2F2) + (num + 1) * 2, SEEK_SET);
		_read(handle, &c, 1);
		next_address = c;
		_read(handle, &c, 1);
		next_address += c * 256;

		_lseek(handle, headerAdjust(0x00D43C) + (num + 1) * 2, SEEK_SET);
		_read(handle, &c, 1);

		next_address += c * 65536;
		next_address = mem2ROM(next_address);
		next_address -= headerAdjust(0x150200);

		size = next_address - address;
	}

	if (type == STATIC_16)
	{
		if (size > 0x80 && num != 152 && num < 162)
			type = STATIC_32;
	}

	tileOffset = address;

	ptr = &charTiles[address];

	tiles.resize(getNumTiles() * 64);

	for (i = 0; i < getNumTiles(); i++)
	{
		loadTile(ptr, &tiles[i * 64]);
		ptr += 32;
	}

	createBitmaps();
}

//Loads the graphics for this map sprite
void Character::load(BYTE* ptr)
{
	for (int i = 0; i < getNumTiles(); i++)
	{
		loadTile(ptr, &tiles[i * 64]);
		ptr += 32;
	}

	createBitmaps();
}

/*************************************/
//Saves the tileset and palette number used by character to the ROM
void Character::save(int handle, BYTE num)
{
	int i, x, y;
	std::deque<bool> tilesDone;
	int offset;
	const USHORT *ptr;
	std::vector<BYTE> saveTiles;
	tilesDone.resize(getNumTiles());
	saveTiles.resize(getNumTiles() * 32);

	if (num < 22)
	{
		_lseek(handle, headerAdjust(0x2D02B) + num, SEEK_SET);

		_write(handle, &palette, 1);
	}

	for (i = 0; i < getNumTiles(); i++)
		tilesDone[i] = false;

	ptr = getTileOffsets();

	for (i = 0; i < getNumPics(); i++)
	{
		for (y = 0; y < getTileHeight(); y++)
		{
			for (x = 0; x < getTileWidth(); x++)
			{
				offset = *(ptr++);
				if (!tilesDone[offset >> 5])
				{
					createTile((pics[i]), x, y, &saveTiles[offset]);
					tilesDone[offset >> 5] = true;
				}
			}
		}
	}

	_lseek(handle, headerAdjust(0x150200) + tileOffset, SEEK_SET);
	for (i = 0; i < getNumTiles(); i++)
	{
		if (tilesDone[i])
			_write(handle, &saveTiles[i * 32], 32);
		else
			_lseek(handle, 32, SEEK_CUR);
	}
}

/*************************************/
//Saves the tileset and palette number used by character to the ROM
void Character::save(int handle)
{
	int i, x, y;
	std::deque<bool> tilesDone;
	int offset;
	const USHORT *ptr;
	std::vector<BYTE> saveTiles;

	tilesDone.resize(getNumTiles(), false);
	saveTiles.resize(getNumTiles() * 32);

	ptr = getTileOffsets();

	for (i = 0; i < getNumPics(); i++)
	{
		for (y = 0; y < getTileHeight(); y++)
		{
			for (x = 0; x < getTileWidth(); x++)
			{
				offset = *(ptr++);
				if (!tilesDone[offset >> 5])
				{
					createTile((pics[i]), x, y, &saveTiles[offset]);
					tilesDone[offset >> 5] = true;
				}
			}
		}
	}

	for (i = 0; i < getNumTiles(); i++)
	{
		if (tilesDone[i])
			_write(handle, &saveTiles[i * 32], 32);
		else
			_lseek(handle, 32, SEEK_CUR);
	}
}

/*************************************/
//Creates the character's sprites from the Bitmap passed in
void Character::setImage(Bitmap& bmp)
{
	int i, j;
	bool found;
	int x, y;
	int x2, y2;
	int sprite;
	RGBTRIPLE color;


	tiles.clear();
	tiles.resize(getNumTiles() * 64);
	const USHORT *offsets = getTileOffsets();
	const int tileCount = getTileWidth() * getTileHeight();
	const int offsetsLength = getNumPics() * tileCount;
	for (i = 0; i < getNumTiles(); i++)
	{
		found = false;
		j = 0;
		do {
			if (offsets[j++] == (i * 32))
			{
				found = true;
			}
		} while ((!found) && (j < offsetsLength));
		if (found == false)
		{
			continue;
		}
		j--;

		sprite = (j / tileCount);
		int width = sprite == numFullPics - 1 ? getTileHeight() : getTileWidth();
		x2 = (sprite % 8) * getPixelWidth();
		x2 += (j % width) * 8;
		y2 = (sprite / 8) * getPixelHeight();
		y2 += ((j - (j / tileCount) * tileCount) / width) * 8;
		y2 += sprite == numFullPics - 1 ? 8 : 0;

		for (y = 0; y < 8; y++)
		{
			for (x = 0; x < 8; x++)
			{
				tiles[i * 64 + x + y * 8] = bmp.getPixel(x + x2, y + y2);
			}
		}
	}

	for (i = 0; i < 16; i++)
	{
		color = bmp.getColor(i);
		charPals[palette][i].rgbtRed = color.rgbtRed;
		charPals[palette][i].rgbtGreen = color.rgbtGreen;
		charPals[palette][i].rgbtBlue = color.rgbtBlue;
	}

	createBitmaps();
}

/*************************************/
//Sets pixcel x,y in the ith sprite for the character
//Updates the coresponding tile
void Character::setPixel(BYTE num, int x, int y, BYTE c)
{
	const USHORT *offsets = getTileOffsets();
	int width = num == numFullPics - 1 ? getTileHeight() : getTileWidth();
	tiles[(offsets[num * 6 + (y / 8) * width + (x / 8)] * 2) + (y % 8) * 8 + (x % 8)] = c;

	createBitmaps();
}

/*************************************/
//Makes color # num transparent by swapping it with color 0
void Character::setTransparency(BYTE num)
{
	RGBTRIPLE temp;
	int i;

	int n = getNumPics();

	if (fixedPalette)
		return;

	for (i = 0; i < n; i++)
		pics[i].swapPal(0, num);

	temp = charPals[palette][0];

	charPals[palette][0].rgbtBlue = charPals[palette][num].rgbtBlue;
	charPals[palette][0].rgbtGreen = charPals[palette][num].rgbtGreen;
	charPals[palette][0].rgbtRed = charPals[palette][num].rgbtRed;
	charPals[palette][num].rgbtBlue = temp.rgbtBlue;
	charPals[palette][num].rgbtGreen = temp.rgbtGreen;
	charPals[palette][num].rgbtRed = temp.rgbtRed;
}

/*************************************/
//Updates the Bitmap with the current palette info
void Character::updatePalette(void)
{
	int i, j;

	if (fixedPalette)
		return;

	for (i = 0; i < 40; i++)
	{
		for (j = 0; j < 16; j++)
		{
			pics[i].setColor(j, charPals[palette][j]);
		}
	}
}

int Character::getTileHeight() const
{
	switch (type)
	{
		case STATIC_32:
			return 4;

		case STATIC_16:
			return 2;

		default:
			return 3;
	}

	return 3;
}

int Character::getTileWidth() const
{
	if (type == STATIC_32)
		return 4;

	return 2;
}

int Character::getNumPics() const
{
	switch (type)
	{
		case FULL:
			return numFullPics;

		case WALK_WAVE_2_POSE:
			return numWalkPics + 4;

		case OPERA:
			return numWalkPics + 3;

		case WALK_POSE:
			return numWalkPics + 1;

		case WALK:
			return numWalkPics;

		case RAMUH:
			return 3;

		case GUARD_RIDING:
			return 2;

		case STATIC_NPC:
			return 1;

		case STATIC_16:
			return size / 0x80;

		case STATIC_32:
			return 1;

		case ANIMATED:
			return (size - 0x20) / 0x80;

		default:
			return numFullPics;
	}

}

bool Character::canCopyFrom(const Character& other, int sprite) const
{
	if (getTileHeight() != other.getTileHeight() || getTileWidth() != other.getTileWidth())
		return false;

	if (type != other.type)
	{
		if (sprite == -1)
			return false;
		bool compatible = (type == FULL && other.type == NO_RIDING ||
			type == NO_RIDING && other.type == FULL);
		if (!compatible)
			return false;
	}

	if (sprite == -1)
	{
		if (type != other.type || getNumPics() != other.getNumPics())
			return false;
	}
	else if (sprite >= getNumPics() || sprite >= other.getNumPics())
		return false;

	if ((type == NO_RIDING || other.type == NO_RIDING) && sprite >= numPreRidingPics && sprite != numFullPics - 1)
		return false;

	return true;
}

/*************************************/
//Loads the sprites for all characters for the ROM
//filename is the name of the ROM to load from
bool loadCharacters(char *filename)
{
	int handle = _open(filename, O_RDWR | O_BINARY);

	if (handle == -1)
		return false;

	const int TILES_PER_PIC = 6;
	const int USHORT_SIZE = sizeof(USHORT);
	const int BYTES_PER_PIC = TILES_PER_PIC * USHORT_SIZE;

	USHORT tileBuffer[115 * TILES_PER_PIC];

	_lseek(handle, headerAdjust(0x0D03A), SEEK_SET);
	_read(handle, tileBuffer, 114 * BYTES_PER_PIC);
	tileBuffer[114 * TILES_PER_PIC] = 0x0AE0;
	tileBuffer[114 * TILES_PER_PIC + 1] = 0x0B00;
	tileBuffer[114 * TILES_PER_PIC + 2] = 0x0B20;
	tileBuffer[114 * TILES_PER_PIC + 3] = 0x0B40;
	tileBuffer[114 * TILES_PER_PIC + 4] = 0x0B60;
	tileBuffer[114 * TILES_PER_PIC + 5] = 0x0B80;

	memcpy(fullOffset, tileBuffer, numPreRidingPics * BYTES_PER_PIC);
	memcpy(&fullOffset[numPreRidingPics * TILES_PER_PIC], &tileBuffer[(numPreRidingPics + 6) * TILES_PER_PIC], 2 * BYTES_PER_PIC);
	memcpy(&fullOffset[(numPreRidingPics + 2) * TILES_PER_PIC], &tileBuffer[114 * TILES_PER_PIC], BYTES_PER_PIC);

	memcpy(walkWave2PoseOffset, tileBuffer, numWalkPics * BYTES_PER_PIC);
	memcpy(operaOffset, tileBuffer, numWalkPics * BYTES_PER_PIC);
	memcpy(walkPoseOffset, tileBuffer, numWalkPics * BYTES_PER_PIC);
	memcpy(walkOffset, tileBuffer, numWalkPics * BYTES_PER_PIC);

	memcpy(&walkWave2PoseOffset[numWalkPics * TILES_PER_PIC], &tileBuffer[0x2A * TILES_PER_PIC], 3 * BYTES_PER_PIC);
	memcpy(&walkWave2PoseOffset[(numWalkPics + 3) * TILES_PER_PIC], &tileBuffer[0x29 * TILES_PER_PIC], BYTES_PER_PIC);

	memcpy(&operaOffset[numWalkPics * TILES_PER_PIC], &tileBuffer[0x36 * TILES_PER_PIC], 2 * BYTES_PER_PIC);
	memcpy(&operaOffset[(numWalkPics + 2) * TILES_PER_PIC], &tileBuffer[0x2D * TILES_PER_PIC], BYTES_PER_PIC);

	memcpy(&walkPoseOffset[numWalkPics * TILES_PER_PIC], &tileBuffer[0x2D * TILES_PER_PIC], BYTES_PER_PIC);

	loadCharPals(handle);
	loadCharTiles(handle);

	for (int i = 0; i < numCharacters; i++)
	{
		character[i].load(handle, i);
	}

	_close(handle);

	return true;
}

/*************************************/
//Loads all 32 palettes used for character sprites
void loadCharPals(int handle)
{
	int i, j;
	std::vector<BYTE> data;
	BYTE *ptr;
	BYTE c1, c2;

	data.reserve(1024);

	_lseek(handle, headerAdjust(0x268200), SEEK_SET);

	_read(handle, data.data(), 1024);

	ptr = data.data();

	for (i = 0; i < 32; i++)
	{
		for (j = 0; j < 16; j++)
		{
			c1 = *(ptr++);
			c2 = *(ptr++);
			charPals[i][j] = getRGBTRIPLE(c1, c2);
		}
	}
}

/*************************************/
//Loads the tiles used for character sprites
void loadCharTiles(int handle)
{
	charTiles.resize(0x35000);

	_lseek(handle, headerAdjust(0x150200), SEEK_SET);

	_read(handle, charTiles.data(), 0x35000);
}


/*************************************/
//Saves all the character sprites to the ROM
bool saveCharacters(char *filename)
{
	int handle;
	int i;

	handle = _open(filename, O_RDWR | O_BINARY);

	if (handle == -1)
	{
		MessageBox(NULL, "Error saving character sprites data", ff3seTitle, MB_OK | MB_ICONWARNING | MB_TASKMODAL);
		return false;
	}

	saveCharPals(handle);

	for (i = 0; i < numCharacters; i++)
	{
		character[i].save(handle, i);
	}

	_close(handle);

	return true;
}

/*************************************/
//Saves all 32 Character sprite palettes to the ROM
void saveCharPals(int handle)
{
	int i, j;
	BYTE c1, c2;

	_lseek(handle, headerAdjust(0x268200), SEEK_SET);

	for (i = 0; i < 32; i++)
	{
		for (j = 0; j < 16; j++)
		{
			getIntPal(charPals[i][j], &c1, &c2);
			_write(handle, &c1, 1);
			_write(handle, &c2, 1);
		}
	}

	_lseek(handle, headerAdjust(0x2D6500), SEEK_SET);

	for (i = 0; i < 6; i++)
	{
		for (j = 0; j < 16; j++)
		{
			getIntPal(charPals[i][j], &c1, &c2);
			_write(handle, &c1, 1);
			_write(handle, &c2, 1);
		}
	}

}

